﻿/*
 *         Progetto: Problema dei produttori/consumatori (buffer singolo)
 * 
 *           Autore: Roberto Fuligni
 *  Ultima modifica: 1/2/2022
 *  
 *      Descrizione: Un produttore e un consumatore si scambiano informazioni
 *                   utilizzando un buffer condiviso a singola cella di memoria.
 *  
 */

using System;
using System.Diagnostics;
using System.Threading;
using System.Threading.Tasks;

namespace BufferSingolo
{
    class BufferSingolo
    {
        const int NumDati = 10;
        const int TempoInserimentoBuffer = 200;    // [ms]
        const int TempoEstrazioneBuffer = 500;     // [ms]
        static SemaphoreSlim vuoto = new SemaphoreSlim(1);
        static SemaphoreSlim pieno = new SemaphoreSlim(0);
        static int buffer;
        static void Produci(int inizio, int fine)
        {
            Stopwatch swProd = new Stopwatch();
            swProd.Start();
            for (int i = inizio; i <= fine; i++)
            {
                int threadId = Thread.CurrentThread.ManagedThreadId;
                int quad = i * i;
                vuoto.Wait();
                Thread.Sleep(TempoInserimentoBuffer);
                buffer = quad;
                pieno.Release();
                Console.WriteLine("Il thread produttore n. {0} ha inserito il numero {1}", threadId, quad);
            }
            swProd.Stop();
            Console.WriteLine("Tempo impiegato dal produttore: {0} ms", swProd.ElapsedMilliseconds);
        }
        static void Consuma(int quanti)
        {
            Stopwatch swCons = new Stopwatch();
            swCons.Start();
            for (int i = 1; i <= quanti; i++)
            {
                int threadId = Thread.CurrentThread.ManagedThreadId;
                int quad;
                pieno.Wait();
                Thread.Sleep(TempoEstrazioneBuffer);
                quad = buffer;
                vuoto.Release();
                Console.WriteLine("Il thread consumatore n. {0} ha estratto il numero {1}", threadId, quad);
            }
            swCons.Stop();
            Console.WriteLine("Tempo impiegato dal consumatore: {0} ms", swCons.ElapsedMilliseconds);
        }
        static void Main(string[] args)
        {
            Console.WriteLine("Un produttore, un consumatore, buffer singolo");
            Console.WriteLine();
            Stopwatch sw = new Stopwatch();
            sw.Start();
            Parallel.Invoke(
                () => Produci(1, NumDati), 
                () => Consuma(NumDati)
            );
            sw.Stop();
            Console.WriteLine("Elaborazione terminata. Tempo totale trascorso: {0} ms", sw.ElapsedMilliseconds);
        }
    }
}
