﻿/*
 *         Progetto: Problema dei produttori/consumatori (produttori e consumatori multipli)
 * 
 *           Autore: Roberto Fuligni
 *  Ultima modifica: 1/2/2022
 *  
 *      Descrizione: Più produttori e consumatori si scambiano informazioni
 *                   utilizzando un buffer circolare condiviso composto da N celle di memoria.
 *  
 */

using System;
using System.Diagnostics;
using System.Threading;
using System.Threading.Tasks;

namespace ProdConsMultipli
{
    class ProdConsMultipli
    {
        const int NumDati = 10;
        const int DimBuffer = 4;
        const int TempoInserimentoBuffer = 200;    // [ms]
        const int TempoEstrazioneBuffer = 500;     // [ms]
        static SemaphoreSlim vuoto = new SemaphoreSlim(DimBuffer);
        static SemaphoreSlim pieno = new SemaphoreSlim(0);
        static SemaphoreSlim mutexP = new SemaphoreSlim(1);
        static SemaphoreSlim mutexC = new SemaphoreSlim(1);
        static int[] buffer = new int[DimBuffer];
        static int metti = 0;
        static int togli = 0;
        static void Produci(int inizio, int fine)
        {
            Stopwatch swProd = new Stopwatch();
            int temp;

            swProd.Start();
            for (int i = inizio; i <= fine; i++)
            {
                int threadId = Thread.CurrentThread.ManagedThreadId;
                int quad = i * i;
                vuoto.Wait();

                mutexP.Wait();
                temp = metti;
                metti = (metti + 1) % DimBuffer;
                mutexP.Release();

                Thread.Sleep(TempoInserimentoBuffer);
                buffer[temp] = quad;
                pieno.Release();
                Console.WriteLine("Il thread produttore n. {0} ha inserito il numero {1}", threadId, quad);
            }
            swProd.Stop();
            Console.WriteLine("Tempo impiegato dal produttore: {0} ms", swProd.ElapsedMilliseconds);
        }
        static void Consuma(int quanti)
        {
            Stopwatch swCons = new Stopwatch();
            int temp;

            swCons.Start();
            for (int i = 1; i <= quanti; i++)
            {
                int threadId = Thread.CurrentThread.ManagedThreadId;
                int quad;
                pieno.Wait();

                mutexC.Wait();
                temp = togli;
                togli = (togli + 1) % DimBuffer;
                mutexC.Release();

                Thread.Sleep(TempoEstrazioneBuffer);
                quad = buffer[temp];
                vuoto.Release();
                Console.WriteLine("Il thread consumatore n. {0} ha estratto il numero {1}", threadId, quad);
            }
            swCons.Stop();
            Console.WriteLine("Tempo impiegato dal consumatore: {0} ms", swCons.ElapsedMilliseconds);
        }
        static void Main(string[] args)
        {
            Console.WriteLine("Due produttori, due consumatori, buffer circolare a {0} celle", DimBuffer);
            Console.WriteLine();
            Stopwatch sw = new Stopwatch();
            sw.Start();
            Parallel.Invoke(
                () => Produci(1, NumDati / 2),
                () => Produci(NumDati / 2 + 1, NumDati),
                () => Consuma(NumDati / 2),
                () => Consuma(NumDati - (NumDati / 2))
            );
            sw.Stop();
            Console.WriteLine("Elaborazione terminata. Tempo trascorso: {0} ms", sw.ElapsedMilliseconds);
        }
    }
}
